/*
 * FAT FS support
 * 
 * Copyright (c) 2021 Ant-ON
 *
 */

#include <sys/param.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>

#include <fs.h>
#include <bm.h>

#include <ext4.h>
#include <ext4_fs.h>
#include <ext4_inode.h>

#define DEBUG_LIB (1ul << 20)

void ext4_crc32_init();

typedef struct
{
	struct ext4_mountpoint *mp;
	struct ext4_blockdev *dev;
} wext_int_t;

static ssize_t wext_f_read(void *vfd, void *buf, size_t len)
{
	fd_t *fd = (fd_t *)vfd;
	ext4_file *file = (ext4_file*)fd->data;

	size_t rcnt;
	int r = ext4_fread(file, buf, len, &rcnt);
    if (r)
		return -1;
	return rcnt;
}

static ssize_t wext_f_write(void *vfd, const void *buf, size_t len)
{
	fd_t *fd = (fd_t *)vfd;
	ext4_file *file = (ext4_file*)fd->data;

	size_t wcnt;
	int r = ext4_fwrite(file, buf, len, &wcnt);
    if (r)
		return -1;
	return wcnt;
}

static off_t wext_f_seek(void *vfd, off_t offset, int dir)
{
	fd_t *fd = (fd_t *)vfd;
	ext4_file *file = (ext4_file*)fd->data;

    ext4_fseek(file, offset, dir);
    return (off_t)ext4_ftell(file);
}

static void wext_f_close(void *vfd)
{
	fd_t *fd = (fd_t *)vfd;
	if (!fd->data)
		return;

	ext4_file *file = (ext4_file*)fd->data;
	fd->data = NULL;

	ext4_fclose(file);
	free(file);
	//free(fd); - free in FS.c
}

static void* wext_fs_fopen(void *vfs, const char *path, int flags)
{
	fs_t *fs = (fs_t *)vfs;
	wext_int_t *in = (wext_int_t*)fs->data;

	ext4_file *file = malloc(sizeof(ext4_file));

	if (ext4_fopen2(file, path, flags, in->mp))
	{
		free(file);
		return NULL;
	}
	
	fd_t *fd = malloc(sizeof(fd_t));
	fd->read = wext_f_read;
	fd->write = wext_f_write;
	fd->seek = wext_f_seek;
	fd->close = wext_f_close;
	fd->data = file;
	return fd;
}

static int wext_fs_mkdir(void *vfs, const char *path)
{
	fs_t *fs = (fs_t *)vfs;
	wext_int_t *in = (wext_int_t*)fs->data;

	return ext4_dir_mk(path, in->mp);
}

static int wext_fs_rename(void *vfs, const char *path, const char *pathNew)
{
	fs_t *fs = (fs_t *)vfs;
	wext_int_t *in = (wext_int_t*)fs->data;

	if (ext4_inode_exist(path, EXT4_DE_DIR, in->mp))
		return ext4_frename(path, pathNew, in->mp);
	else
		return ext4_dir_mv(path, pathNew, in->mp);
}

static int wext_fs_remove(void *vfs, const char *path)
{
	fs_t *fs = (fs_t *)vfs;
	wext_int_t *in = (wext_int_t*)fs->data;

	if (ext4_inode_exist(path, EXT4_DE_DIR, in->mp))
		return ext4_fremove(path, in->mp);
	else
		return ext4_dir_rm(path, in->mp);
}

static int wext_fs_list(void *vfs, void *ctx, const char *path, _filter filter)
{
	fs_t *fs = (fs_t *)vfs;
	wext_int_t *in = (wext_int_t*)fs->data;

	ext4_dir dir;
    int r = ext4_dir_open(&dir, path, in->mp);
    if (!r)
    {
		ext4_direntry *de = ext4_dir_entry_next(&dir);
		while (de)
		{
			int32_t mtime;
			struct ext4_inode_ref inode_ref;
			
			if ((de->name_length == 1 && de->name[0] == '.') ||
				(de->name_length == 2 && de->name[0] == '.' && de->name[1] == '.'))
			{
				de = ext4_dir_entry_next(&dir);
				continue;
			}

			/*if (!ext4_fs_get_inode_ref(in->dev->fs, de->inode, &inode_ref))
			{
				mtime = ext4_inode_get_modif_time(inode_ref.inode);
				ext4_fs_put_inode_ref(&inode_ref);
			} else*/
				mtime = 0;

			size_t len = de->name_length >= 255 ? 254 : de->name_length;
			de->name[len] = '\0';

			int flags = (de->inode_type&EXT4_DE_DIR)?F_FLAG_DIR:0;
			filter(ctx, de->name, de->entry_length, mtime, flags);

			de = ext4_dir_entry_next(&dir);
		}

		ext4_dir_close(&dir);
	}

	return r;
}

static int wext_fs_stat(void *vfs, const char *path, stat_t *stat)
{
	fs_t *fs = (fs_t *)vfs;
	wext_int_t *in = (wext_int_t*)fs->data;

	stat->time = 0;

	ext4_file file;
    int r = ext4_fopen2(&file, path, O_RDONLY, in->mp);
	if (!r)
	{
		ext4_mtime_get(NULL, &file, (uint32_t*)&stat->time, in->mp);

		stat->size = ext4_fsize(&file);
		stat->flags = 0;
		ext4_fclose(&file);
	} else
	{
		r = ext4_inode_exist(path, EXT4_DE_DIR, in->mp);
		stat->flags = F_FLAG_DIR;
	}

	return r;
}

static int wext_fs_statfs(void *vfs, stat_fs_t *stat_fs)
{
	fs_t *fs = (fs_t *)vfs;
	wext_int_t *in = (wext_int_t*)fs->data;

	struct ext4_mount_stats stats;
	int r = ext4_mount_point_stats(in->mp, &stats);
	if (!r)
	{
		stat_fs->name = strdup(stats.volume_name);
		stat_fs->size = (int64_t)stats.blocks_count * stats.block_size;
		stat_fs->free = (int64_t)stats.free_blocks_count * stats.block_size;
		stat_fs->flags = 0;
	}

	return r;
}

static void wext_fs_fsfree(void *vfs)
{
	fs_t *fs = (fs_t *)vfs;
	wext_int_t *in = (wext_int_t*)fs->data;

	if (in)
	{
		ext4_cache_write_back(false, in->mp);
		ext4_journal_stop(in->mp);
		ext4_umount(in->mp);

		if (in->dev)
		{
			if (in->dev->bdif)
			{
				if (in->dev->bdif->ph_bbuf)
					free(in->dev->bdif->ph_bbuf);
				free(in->dev->bdif);
			}
			free(in->dev);
		}
		free(in);
	}
	
	free(fs);
}

static int bd_wext_open(struct ext4_blockdev *bdev)
{
    return EOK;
}

static int bd_wext_bread(struct ext4_blockdev *bdev, void *buf, uint64_t blk_id, uint32_t blk_cnt)
{
	block_t *blk = (block_t*)bdev->bdif->p_user;
	if (!blk)
		return EIO;

	ssize_t ret = blk->pread(blk, buf, (size_t)blk_cnt*bdev->bdif->ph_bsize, (off_t)blk_id*bdev->bdif->ph_bsize);
    if (ret < 0)
        return EIO;
    return EOK;
}

static int bd_wext_bwrite(struct ext4_blockdev *bdev, const void *buf, uint64_t blk_id, uint32_t blk_cnt)
{
	block_t *blk = (block_t*)bdev->bdif->p_user;
	if (!blk)
		return EIO;

	ssize_t ret = blk->pwrite(blk, buf, (size_t)blk_cnt*bdev->bdif->ph_bsize, (off_t)blk_id*bdev->bdif->ph_bsize);
    if (ret < 0)
        return EIO;
    return EOK;
}

static int bd_wext_close(struct ext4_blockdev *bdev)
{
    (void)bdev;
    return EOK;
}

#ifdef MODULE_FS
void* __UI fs_open(void *block, int flags)
#else
void* __UI wext_fs_open(void *block, int flags)
#endif
{
	block_t *blk = (block_t*)block;
	struct ext4_blockdev_iface* iface = NULL;
	struct ext4_blockdev *dev = NULL;

	uint32_t sector_size;
	if (blk->ioctl(blk, IOCTL_GET_SECTOR_SIZE, &sector_size))
	{
		ext4_dbg(DEBUG_LIB, "[FS] IOCTL_GET_SECTOR_SIZE failed");
		goto error;
	}
	uint32_t sector_count;
	if (blk->ioctl(blk, IOCTL_GET_SECTOR_COUNT, &sector_count))
	{
		ext4_dbg(DEBUG_LIB, "[FS] IOCTL_GET_SECTOR_COUNT failed");
		goto error;
	}

	ext4_crc32_init();

    iface = (struct ext4_blockdev_iface*)malloc(sizeof(struct ext4_blockdev_iface));          
    iface->open = bd_wext_open;
    iface->bread = bd_wext_bread;
    iface->bwrite = bd_wext_bwrite;
    iface->close = bd_wext_close;
    iface->lock = NULL;
    iface->unlock = NULL;
    iface->ph_bsize = sector_size;
    iface->ph_bcnt = sector_count;
    iface->ph_bbuf = (uint8_t *)malloc(sector_size);
	iface->ph_refctr = 0;
	iface->bread_ctr = 0;
	iface->bwrite_ctr = 0;
	iface->p_user = blk;

	dev = (struct ext4_blockdev *)malloc(sizeof(struct ext4_blockdev));
    dev->bdif = iface;
    dev->part_offset = 0;
    dev->part_size = sector_size * sector_count;
	dev->bc = NULL;
	dev->lg_bsize = sector_size;
	dev->lg_bcnt = sector_count;
	dev->cache_write_back = 0;
	dev->fs = NULL;
	dev->journal = NULL;

	struct ext4_mountpoint *mount_point;
    if (ext4_mount(dev, &mount_point, !!(flags&FLAG_FS_OPEN_RO), NULL))
	{
		ext4_dbg(DEBUG_LIB, "[FS] device mount failed");
		goto error;
	}

	ext4_recover(mount_point);
	ext4_journal_start(mount_point);
	ext4_cache_write_back(true, mount_point);

	fs_t *fs = malloc(sizeof(fs_t));
	fs->open = wext_fs_fopen;
	fs->mkdir = wext_fs_mkdir;
	fs->rename = wext_fs_rename;
	fs->remove = wext_fs_remove;
	fs->list = wext_fs_list;
	fs->stat = wext_fs_stat;
	fs->statfs = wext_fs_statfs;
	fs->fsfree = wext_fs_fsfree;
	fs->data = malloc(sizeof(wext_int_t));
	((wext_int_t*)fs->data)->mp = mount_point;
	((wext_int_t*)fs->data)->dev = dev;
	return fs;

error:
	if (iface)
	{
		if (iface->ph_bbuf)
			free(iface->ph_bbuf);
		free(iface);
	}
	if (dev)
		free(dev);
	return NULL;
}
